/*
 *  Copyright (C) 2007  Anders Gavare.  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright  
 *     notice, this list of conditions and the following disclaimer in the 
 *     documentation and/or other materials provided with the distribution.
 *  3. The name of the author may not be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE   
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 *
 *
 *  $Id: native.c,v 1.2 2007/02/16 16:48:07 debug Exp $
 *
 *  Native Code Generation framework; common routines.
 *
 *  NOTE/TODO:  This is just a dummy so far, native code generation
 *              has not been implemented.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "cpu.h"
#include "native.h"


#ifndef NATIVE_CODE_GENERATION


/*
 *  Dummy functions, if native code generation is not supported
 *  for this specific host.
 */
void native_start(struct cpu *cpu) { }
void native_commit(struct cpu *cpu) { }


#else	/*  NATIVE_CODE_GENERATION  */


#ifdef HOST_ARCH_AMD64
#include "native_amd64.c"
#endif

extern int native_code_translation_enabled;


/*
 *  native_start():
 *
 *  This function should be the first thing called in each individual
 *  instruction output function.
 */
void native_start(struct cpu *cpu)
{
	if (cpu->currently_translating_to_native)
		return;

	cpu->currently_translating_to_native = 1;
	cpu->nr_of_instructions_translated_to_native = 0;
	cpu->native_cur_output_ptr = cpu->translation_cache
	    + cpu->translation_cache_cur_ofs;

	native_output_function_prelude(cpu);
}


/*
 *  native_commit():
 *
 *  This function should be called when the current translation (if any)
 *  is to be committed to its corresponding ic->f pointer.
 *
 *  It updates the current cache offset (cpu->translation_cache_cur_ofs).
 */
void native_commit(struct cpu *cpu)
{
	if (!native_code_translation_enabled ||
	    !cpu->currently_translating_to_native)
		return;

	cpu->currently_translating_to_native = 0;

	native_output_next_ic_increment(cpu);
	native_output_function_postlude(cpu);

#if 0
{
uint8_t *p = cpu->translation_cache + cpu->translation_cache_cur_ofs;
while (p < cpu->native_cur_output_ptr)
	printf(" %02x", *p++);
printf("\n");
}
#endif

	(*cpu->native_code_function_pointer) = (void *)
	    (cpu->translation_cache + cpu->translation_cache_cur_ofs);

	cpu->translation_cache_cur_ofs = cpu->native_cur_output_ptr
	    - cpu->translation_cache;

	cpu->translation_cache_cur_ofs --;
	cpu->translation_cache_cur_ofs |= (NATIVE_FUNCTION_ALIGNMENT - 1);
	cpu->translation_cache_cur_ofs ++;
}


#endif	/*  NATIVE_CODE_GENERATION  */

