/*
 *  Copyright (C) 2007  Anders Gavare.  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright  
 *     notice, this list of conditions and the following disclaimer in the 
 *     documentation and/or other materials provided with the distribution.
 *  3. The name of the author may not be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE   
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 *
 *
 *  $Id: native_amd64.c,v 1.2 2007/02/16 16:48:08 debug Exp $
 *
 *  Native code generation backend for AMD64 and i386.
 *
 *  TODO: This is just a dummy so far, there is no native code generation
 *        in GXemul.
 *
 *  TODO 2: This is not really for i386 (32-bit) yet.
 *
 *  HOST_ARCH_AMD64 is set for AMD64 hosts, and HOST_ARCH_I386 for i386 hosts.
 *
 *  Argument order:  edi, esi, edx, ecx, e8d, r9d, ...
 *  edi should always point to the cpu struct, esi to the ic.
 */

#define	NATIVE_FUNCTION_ALIGNMENT		32

#define	OUTPUT_32(v) {						\
	*p++ = (v) & 255; *p++ = ((v) >> 8) & 255;		\
	*p++ = ((v) >> 16) & 255; *p++ = ((v) >> 24) & 255; }
#define	OUTPUT_64(v) {						\
	*p++ = (v) & 255; *p++ = ((v) >> 8) & 255;		\
	*p++ = ((v) >> 16) & 255; *p++ = ((v) >> 24) & 255;	\
	*p++ = ((v) >> 32) & 255; *p++ = ((v) >> 40) & 255;	\
	*p++ = ((v) >> 48) & 255; *p++ = ((v) >> 56) & 255; }
#ifdef HOST_ARCH_AMD64
#define	OUTPUT_HOST_SIZE(v)	OUTPUT_64(v)
#else
#define	OUTPUT_HOST_SIZE(v)	OUTPUT_32(v)
#endif

void native_output_function_prelude(struct cpu *cpu)
{
	printf("no\n");
	abort();
}

void native_output_next_ic_increment(struct cpu *cpu)
{
	uint8_t *p = cpu->native_cur_output_ptr;
	uint64_t y1 = (size_t) (void *) &cpu->n_translated_instrs;
	uint64_t y2 = (size_t) (void *) &cpu->cd.mips.next_ic;
	int inc = cpu->nr_of_instructions_translated_to_native - 1;

	printf("not yet\n");
	abort();

	*p++ = 0x48; *p++ = 0xb8;
	OUTPUT_HOST_SIZE(y1);		/*  rax = y1  */

	*p++ = 0x48; *p++ = 0xba;
	OUTPUT_HOST_SIZE(y2);		/*  rdx = y2  */

	*p++ = 0x81; *p++ = 0x00;	/*  add 32-bit word to (rax)  */
	OUTPUT_32(inc);

	inc *= 0x20;

	*p++ = 0x48; *p++ = 0x81; *p++ = 0x02;	/*  add 64-bit word to (rdx)  */
	OUTPUT_32(inc);

	cpu->native_cur_output_ptr = p;
}

void native_output_function_postlude(struct cpu *cpu)
{
	uint8_t *p = cpu->native_cur_output_ptr;
	*p++ = 0xc3;	/*  ret, or retq  */
	cpu->native_cur_output_ptr = p;
}

int native_nop(struct cpu *cpu)
{
	native_start(cpu);
	cpu->nr_of_instructions_translated_to_native ++;
	return 1;
}

int native_set_u32_p32(struct cpu *cpu, uint32_t v, uint32_t *p1)
{
	uint64_t y1 = (size_t) (void *) p1;
	uint8_t *p;

	native_start(cpu);

	p = cpu->native_cur_output_ptr;

	*p++ = 0xb8;		/*  set eax to value  */
	OUTPUT_32(v);

	*p++ = 0xa3;		/*  store eax into memory  */
	OUTPUT_HOST_SIZE(y1);

	cpu->nr_of_instructions_translated_to_native ++;
	cpu->native_cur_output_ptr = p;

	return 1;
}

